# @Author: Bi Ying
# @Date:   2024-07-26 14:48:55
from collections.abc import Generator, AsyncGenerator, Iterable
from typing import Any, TYPE_CHECKING, overload, Literal, cast

import httpx
from openai._types import NotGiven as OpenAINotGiven
from openai._types import NOT_GIVEN as OPENAI_NOT_GIVEN
from openai._types import Headers, Query, Body
from openai.types.shared_params.metadata import Metadata
from openai.types.chat.completion_create_params import ResponseFormat
from openai.types.chat.chat_completion_modality import ChatCompletionModality
from openai.types.chat.chat_completion_audio_param import ChatCompletionAudioParam
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from openai.types.chat.chat_completion_stream_options_param import ChatCompletionStreamOptionsParam
from openai.types.chat.chat_completion_prediction_content_param import ChatCompletionPredictionContentParam
from anthropic import (
    Anthropic,
    AnthropicVertex,
    AsyncAnthropic,
    AsyncAnthropicVertex,
    AnthropicBedrock,
    AsyncAnthropicBedrock,
)
from anthropic._types import NOT_GIVEN, omit, Omit
from anthropic._exceptions import APIStatusError as AnthropicAPIStatusError
from anthropic._exceptions import APIConnectionError as AnthropicAPIConnectionError
from anthropic.types import (
    ThinkingConfigParam,
)

from ..types import defaults as defs
from .utils import cutoff_messages, get_message_token_counts
from .base_client import BaseChatClient, BaseAsyncChatClient
from .message_normalizer import format_messages_alternate, refactor_into_openai_messages
from .openai_compatible_client import OpenAICompatibleChatClient, AsyncOpenAICompatibleChatClient
from .stream_event_adapter import (
    adapt_anthropic_stream_event,
    build_anthropic_completion_message,
    init_anthropic_stream_state,
)
from .tool_call_parser import refactor_tool_choice, refactor_tool_use_params
from ..types.exception import APIStatusError, APIConnectionError
from ..types.enums import ContextLengthControlType, BackendType
from ..utilities.gcp_token import get_token_with_cache
from ..types.llm_parameters import NotGiven, ToolParam, ToolChoice, AnthropicToolParam, ChatCompletionMessage, ChatCompletionDeltaMessage

if TYPE_CHECKING:
    from ..settings import Settings
    from ..types.settings import SettingsDict


class AnthropicChatClient(BaseChatClient):
    DEFAULT_MODEL: str = defs.ANTHROPIC_DEFAULT_MODEL
    BACKEND_NAME: BackendType = BackendType.Anthropic

    def __init__(
        self,
        model: str = defs.ANTHROPIC_DEFAULT_MODEL,
        stream: bool = True,
        temperature: float | None | NotGiven = NOT_GIVEN,
        context_length_control: ContextLengthControlType = defs.CONTEXT_LENGTH_CONTROL,
        random_endpoint: bool = True,
        endpoint_id: str = "",
        http_client: httpx.Client | None = None,
        backend_name: str | None = None,
        settings: "Settings | SettingsDict | None" = None,  # Use default settings if not provided
    ):
        super().__init__(
            model,
            stream,
            temperature,
            context_length_control,
            random_endpoint,
            endpoint_id,
            http_client,
            backend_name,
            settings,
        )
        self.model_id = None
        self.endpoint = None
        self._cached_raw_client = None

    @property
    def raw_client(self):
        self.endpoint, self.model_id = self._set_endpoint()

        if self.endpoint.proxy and self.http_client is None:
            self.http_client = httpx.Client(proxy=self.endpoint.proxy)

        if self.endpoint.is_vertex or self.endpoint.endpoint_type == "anthropic_vertex":
            # Vertex needs token refresh on each access
            if self.endpoint.credentials is None:
                raise ValueError("Anthropic Vertex endpoint requires credentials")

            access_token, expires_at = get_token_with_cache(
                credentials=self.endpoint.credentials,
                proxy=self.endpoint.proxy,
                cached_token=self.endpoint.access_token,
                cached_expires_at=self.endpoint.access_token_expires_at,
            )
            self.endpoint.access_token = access_token
            self.endpoint.access_token_expires_at = expires_at

            if self.endpoint.api_base is None:
                base_url = None
            else:
                base_url = f"{self.endpoint.api_base}{self.endpoint.region}-aiplatform/v1"

            region = NOT_GIVEN if self.endpoint.region is None else self.endpoint.region
            return AnthropicVertex(
                region=region,
                base_url=base_url,
                project_id=self.endpoint.credentials.get("quota_project_id", NOT_GIVEN),
                access_token=access_token,
                http_client=self.http_client,
            )
        elif self.endpoint.is_bedrock or self.endpoint.endpoint_type == "anthropic_bedrock":
            if self.endpoint.credentials is None:
                raise ValueError("Anthropic Bedrock endpoint requires credentials")
            return AnthropicBedrock(
                aws_access_key=self.endpoint.credentials.get("access_key"),
                aws_secret_key=self.endpoint.credentials.get("secret_key"),
                aws_region=self.endpoint.region,
                base_url=self.endpoint.api_base,
                http_client=self.http_client,
            )
        elif self.endpoint.endpoint_type in ("default", "anthropic"):
            if self._cached_raw_client is None:
                self._cached_raw_client = Anthropic(
                    api_key=self.endpoint.api_key,
                    base_url=self.endpoint.api_base,
                    http_client=self.http_client,
                )
            return self._cached_raw_client
        else:
            if self._cached_raw_client is None:
                self._cached_raw_client = OpenAICompatibleChatClient(
                    model=self.model,
                    stream=self.stream,
                    temperature=cast(Any, self.temperature),
                    context_length_control=self.context_length_control,
                    random_endpoint=self.random_endpoint,
                    endpoint_id=self.endpoint_id,
                    http_client=self.http_client,
                    backend_name=self.BACKEND_NAME,
                ).raw_client
            return self._cached_raw_client

    @overload
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] = False,
        temperature: float | None | NotGiven = OPENAI_NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = OPENAI_NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = OPENAI_NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = OPENAI_NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = OPENAI_NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        n: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = OPENAI_NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = OPENAI_NOT_GIVEN,
        seed: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        store: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        user: str | OpenAINotGiven = OPENAI_NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = OPENAI_NOT_GIVEN,
    ) -> ChatCompletionMessage:
        pass

    @overload
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[True],
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        n: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = OPENAI_NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        store: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        user: str | OpenAINotGiven = OPENAI_NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = OPENAI_NOT_GIVEN,
    ) -> Generator[ChatCompletionDeltaMessage, None, None]:
        pass

    @overload
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: bool,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        n: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = OPENAI_NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        store: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        user: str | OpenAINotGiven = OPENAI_NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = OPENAI_NOT_GIVEN,
    ) -> ChatCompletionMessage | Generator[ChatCompletionDeltaMessage, Any, None]:
        pass

    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] | Literal[True] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        n: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = OPENAI_NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        store: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        user: str | OpenAINotGiven = OPENAI_NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = OPENAI_NOT_GIVEN,
    ):
        if model is not None:
            self.model = model
        if stream is not None:
            self.stream = stream
        if temperature is not None:
            self.temperature = temperature

        self.model_setting = self.backend_settings.models[self.model]
        if self.model_id is None:
            self.model_id = self.model_setting.id

        self.endpoint, self.model_id = self._set_endpoint()

        if self.endpoint.endpoint_type and self.endpoint.endpoint_type.startswith("openai"):
            _tools = OPENAI_NOT_GIVEN if tools is NOT_GIVEN else tools
            _tool_choice = OPENAI_NOT_GIVEN if tool_choice is NOT_GIVEN else tool_choice

            formatted_messages = refactor_into_openai_messages(messages)

            if self.stream:

                def _generator():
                    openai_client = cast(
                        Any,
                        OpenAICompatibleChatClient(
                            model=self.model,
                            stream=True,
                            temperature=cast(Any, self.temperature),
                            context_length_control=self.context_length_control,
                            random_endpoint=self.random_endpoint,
                            endpoint_id=self.endpoint_id,
                            http_client=self.http_client,
                            backend_name=self.BACKEND_NAME,
                        ),
                    )
                    response = openai_client.create_completion(
                        messages=formatted_messages,
                        model=model,
                        stream=True,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        tools=_tools,
                        tool_choice=_tool_choice,
                        response_format=response_format,
                        stream_options=stream_options,
                        top_p=top_p,
                        skip_cutoff=skip_cutoff,
                        audio=audio,
                        frequency_penalty=frequency_penalty,
                        logit_bias=logit_bias,
                        logprobs=logprobs,
                        max_completion_tokens=max_completion_tokens,
                        metadata=metadata,
                        modalities=modalities,
                        n=n,
                        parallel_tool_calls=parallel_tool_calls,
                        prediction=prediction,
                        presence_penalty=presence_penalty,
                        reasoning_effort=reasoning_effort,
                        seed=seed,
                        service_tier=service_tier,
                        stop=stop,
                        store=store,
                        top_logprobs=top_logprobs,
                        user=user,
                        extra_headers=extra_headers,
                        extra_query=extra_query,
                        extra_body=extra_body,
                        timeout=timeout,
                    )
                    yield from response

                return _generator()
            else:
                openai_client = cast(
                    Any,
                    OpenAICompatibleChatClient(
                        model=self.model,
                        stream=False,
                        temperature=cast(Any, self.temperature),
                        context_length_control=self.context_length_control,
                        random_endpoint=self.random_endpoint,
                        endpoint_id=self.endpoint_id,
                        http_client=self.http_client,
                        backend_name=self.BACKEND_NAME,
                    ),
                )
                return openai_client.create_completion(
                    messages=formatted_messages,
                    model=model,
                    stream=False,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    tools=_tools,
                    tool_choice=_tool_choice,
                    response_format=response_format,
                    top_p=top_p,
                    skip_cutoff=skip_cutoff,
                    audio=audio,
                    frequency_penalty=frequency_penalty,
                    logit_bias=logit_bias,
                    logprobs=logprobs,
                    max_completion_tokens=max_completion_tokens,
                    metadata=metadata,
                    modalities=modalities,
                    n=n,
                    parallel_tool_calls=parallel_tool_calls,
                    prediction=prediction,
                    presence_penalty=presence_penalty,
                    reasoning_effort=reasoning_effort,
                    seed=seed,
                    service_tier=service_tier,
                    stop=stop,
                    store=store,
                    top_logprobs=top_logprobs,
                    user=user,
                    extra_headers=extra_headers,
                    extra_query=extra_query,
                    extra_body=extra_body,
                    timeout=timeout,
                )

        raw_client = self.raw_client  # 调用完 self.raw_client 后，self.model_id 会被赋值
        assert isinstance(raw_client, Anthropic | AnthropicVertex | AnthropicBedrock)

        _tools = omit if isinstance(tools, NotGiven) else tools
        _tool_choice = omit if isinstance(tool_choice, NotGiven) else tool_choice
        _top_p = omit if isinstance(top_p, NotGiven) or top_p is None else top_p
        if isinstance(self.temperature, NotGiven) or self.temperature is None:
            self.temperature = omit
        _thinking = omit if isinstance(thinking, NotGiven) or thinking is None else thinking

        if messages[0].get("role") == "system":
            system_content = messages[0]["content"]
            # Preserve list format for system prompt to support cache_control
            if isinstance(system_content, list):
                system_prompt = system_content  # Keep as list to preserve cache_control
            else:
                system_prompt = system_content  # String format
            messages = messages[1:]
        else:
            system_prompt = ""

        if not skip_cutoff and self.context_length_control == ContextLengthControlType.Latest:
            messages = cutoff_messages(
                messages,
                max_count=self.model_setting.context_length,
                backend=self.BACKEND_NAME,
                model=self.model,
            )

        messages = format_messages_alternate(messages)

        tools_params: list[AnthropicToolParam] | Omit = refactor_tool_use_params(_tools) if _tools else omit
        tool_choice_param = omit
        if _tool_choice:
            tool_choice_param = refactor_tool_choice(_tool_choice)

        if not max_tokens:
            max_output_tokens = self.model_setting.max_output_tokens
            native_multimodal = self.model_setting.native_multimodal
            token_counts = get_message_token_counts(messages=messages, tools=tools, model=self.model, native_multimodal=native_multimodal)
            if max_output_tokens is not None:
                max_tokens = self.model_setting.context_length - token_counts
                max_tokens = min(max(max_tokens, 1), max_output_tokens)
            else:
                max_tokens = self.model_setting.context_length - token_counts

        self._acquire_rate_limit(self.endpoint, self.model, messages)

        if self.stream:
            try:
                stream_response = raw_client.messages.create(
                    max_tokens=max_tokens,
                    messages=messages,
                    model=self.model_id,
                    stream=True,
                    system=system_prompt,
                    temperature=self.temperature,
                    tools=tools_params,
                    tool_choice=tool_choice_param,
                    top_p=_top_p,
                    thinking=_thinking,
                )
            except AnthropicAPIStatusError as e:
                raise APIStatusError(message=e.message, response=e.response, body=e.body) from e
            except AnthropicAPIConnectionError as e:
                raise APIConnectionError(message=e.message, request=e.request) from e

            def generator():
                stream_state = init_anthropic_stream_state()
                for chunk in stream_response:
                    message = adapt_anthropic_stream_event(chunk, stream_state)
                    if message is not None:
                        yield message

            return generator()
        else:
            try:
                response = raw_client.messages.create(
                    model=self.model_id,
                    messages=messages,
                    system=system_prompt,
                    stream=False,
                    temperature=self.temperature,
                    max_tokens=max_tokens,
                    tools=tools_params,
                    tool_choice=tool_choice_param,
                    top_p=_top_p,
                    thinking=_thinking,
                )
            except AnthropicAPIStatusError as e:
                raise APIStatusError(message=e.message, response=e.response, body=e.body) from e
            except AnthropicAPIConnectionError as e:
                raise APIConnectionError(message=e.message, request=e.request) from e

            return build_anthropic_completion_message(list(response.content), response.usage)


class AsyncAnthropicChatClient(BaseAsyncChatClient):
    DEFAULT_MODEL: str = defs.ANTHROPIC_DEFAULT_MODEL
    BACKEND_NAME: BackendType = BackendType.Anthropic

    def __init__(
        self,
        model: str = defs.ANTHROPIC_DEFAULT_MODEL,
        stream: bool = True,
        temperature: float | None | NotGiven = NOT_GIVEN,
        context_length_control: ContextLengthControlType = defs.CONTEXT_LENGTH_CONTROL,
        random_endpoint: bool = True,
        endpoint_id: str = "",
        http_client: httpx.AsyncClient | None = None,
        backend_name: str | None = None,
        settings: "Settings | SettingsDict | None" = None,  # Use default settings if not provided
    ):
        super().__init__(
            model,
            stream,
            temperature,
            context_length_control,
            random_endpoint,
            endpoint_id,
            http_client,
            backend_name,
            settings,
        )
        self.model_id = None
        self.endpoint = None
        self._cached_raw_client = None

    @property
    def raw_client(self):
        self.endpoint, self.model_id = self._set_endpoint()

        if self.endpoint.proxy and self.http_client is None:
            self.http_client = httpx.AsyncClient(proxy=self.endpoint.proxy)

        if self.endpoint.is_vertex or self.endpoint.endpoint_type == "anthropic_vertex":
            # Vertex needs token refresh on each access
            if self.endpoint.credentials is None:
                raise ValueError("Anthropic Vertex endpoint requires credentials")

            access_token, expires_at = get_token_with_cache(
                credentials=self.endpoint.credentials,
                proxy=self.endpoint.proxy,
                cached_token=self.endpoint.access_token,
                cached_expires_at=self.endpoint.access_token_expires_at,
            )
            self.endpoint.access_token = access_token
            self.endpoint.access_token_expires_at = expires_at

            if self.endpoint.api_base is None:
                base_url = None
            else:
                base_url = f"{self.endpoint.api_base}{self.endpoint.region}-aiplatform/v1"

            region = NOT_GIVEN if self.endpoint.region is None else self.endpoint.region
            return AsyncAnthropicVertex(
                region=region,
                base_url=base_url,
                project_id=self.endpoint.credentials.get("quota_project_id", NOT_GIVEN),
                access_token=access_token,
                http_client=self.http_client,
            )
        elif self.endpoint.is_bedrock or self.endpoint.endpoint_type == "anthropic_bedrock":
            if self.endpoint.credentials is None:
                raise ValueError("Anthropic Bedrock endpoint requires credentials")
            return AsyncAnthropicBedrock(
                aws_access_key=self.endpoint.credentials.get("access_key"),
                aws_secret_key=self.endpoint.credentials.get("secret_key"),
                aws_region=self.endpoint.region,
                http_client=self.http_client,
            )
        elif self.endpoint.endpoint_type in ("default", "anthropic"):
            if self._cached_raw_client is None:
                self._cached_raw_client = AsyncAnthropic(
                    api_key=self.endpoint.api_key,
                    base_url=self.endpoint.api_base,
                    http_client=self.http_client,
                )
            return self._cached_raw_client
        else:
            if self._cached_raw_client is None:
                self._cached_raw_client = AsyncOpenAICompatibleChatClient(
                    model=self.model,
                    stream=self.stream,
                    temperature=cast(Any, self.temperature),
                    context_length_control=self.context_length_control,
                    random_endpoint=self.random_endpoint,
                    endpoint_id=self.endpoint_id,
                    http_client=self.http_client,
                    backend_name=self.BACKEND_NAME,
                ).raw_client
            return self._cached_raw_client

    @overload
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        n: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = OPENAI_NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        store: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        user: str | OpenAINotGiven = OPENAI_NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = OPENAI_NOT_GIVEN,
    ) -> ChatCompletionMessage:
        pass

    @overload
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[True],
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        n: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = OPENAI_NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        store: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        user: str | OpenAINotGiven = OPENAI_NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = OPENAI_NOT_GIVEN,
    ) -> AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        pass

    @overload
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: bool,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        n: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = OPENAI_NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        store: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        user: str | OpenAINotGiven = OPENAI_NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = OPENAI_NOT_GIVEN,
    ) -> ChatCompletionMessage | AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        pass

    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] | Literal[True] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        n: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = OPENAI_NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        store: bool | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = OPENAI_NOT_GIVEN,
        user: str | OpenAINotGiven = OPENAI_NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = OPENAI_NOT_GIVEN,
    ) -> ChatCompletionMessage | AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        if model is not None:
            self.model = model
        if stream is not None:
            self.stream = stream
        if temperature is not None:
            self.temperature = temperature

        self.model_setting = self.backend_settings.models[self.model]
        if self.model_id is None:
            self.model_id = self.model_setting.id

        self.endpoint, self.model_id = self._set_endpoint()

        if self.endpoint.endpoint_type and self.endpoint.endpoint_type.startswith("openai"):
            _tools = OPENAI_NOT_GIVEN if tools is NOT_GIVEN else tools
            _tool_choice = OPENAI_NOT_GIVEN if tool_choice is NOT_GIVEN else tool_choice

            formatted_messages = refactor_into_openai_messages(messages)

            if self.stream:

                async def _generator():
                    client = cast(
                        Any,
                        AsyncOpenAICompatibleChatClient(
                            model=self.model,
                            stream=True,
                            temperature=cast(Any, self.temperature),
                            context_length_control=self.context_length_control,
                            random_endpoint=self.random_endpoint,
                            endpoint_id=self.endpoint_id,
                            http_client=self.http_client,
                            backend_name=self.BACKEND_NAME,
                        ),
                    )
                    response = await client.create_completion(
                        messages=formatted_messages,
                        model=model,
                        stream=True,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        tools=_tools,
                        tool_choice=_tool_choice,
                        response_format=response_format,
                        stream_options=stream_options,
                        top_p=top_p,
                        skip_cutoff=skip_cutoff,
                        audio=audio,
                        frequency_penalty=frequency_penalty,
                        logit_bias=logit_bias,
                        logprobs=logprobs,
                        max_completion_tokens=max_completion_tokens,
                        metadata=metadata,
                        modalities=modalities,
                        n=n,
                        parallel_tool_calls=parallel_tool_calls,
                        prediction=prediction,
                        presence_penalty=presence_penalty,
                        reasoning_effort=reasoning_effort,
                        seed=seed,
                        service_tier=service_tier,
                        stop=stop,
                        store=store,
                        top_logprobs=top_logprobs,
                        user=user,
                        extra_headers=extra_headers,
                        extra_query=extra_query,
                        extra_body=extra_body,
                        timeout=timeout,
                    )
                    async for chunk in response:
                        yield chunk

                return _generator()
            else:
                client = cast(
                    Any,
                    AsyncOpenAICompatibleChatClient(
                        model=self.model,
                        stream=False,
                        temperature=cast(Any, self.temperature),
                        context_length_control=self.context_length_control,
                        random_endpoint=self.random_endpoint,
                        endpoint_id=self.endpoint_id,
                        http_client=self.http_client,
                        backend_name=self.BACKEND_NAME,
                    ),
                )
                return await client.create_completion(
                    messages=formatted_messages,
                    model=model,
                    stream=False,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    tools=_tools,
                    tool_choice=_tool_choice,
                    response_format=response_format,
                    top_p=top_p,
                    skip_cutoff=skip_cutoff,
                    audio=audio,
                    frequency_penalty=frequency_penalty,
                    logit_bias=logit_bias,
                    logprobs=logprobs,
                    max_completion_tokens=max_completion_tokens,
                    metadata=metadata,
                    modalities=modalities,
                    n=n,
                    parallel_tool_calls=parallel_tool_calls,
                    prediction=prediction,
                    presence_penalty=presence_penalty,
                    reasoning_effort=reasoning_effort,
                    seed=seed,
                    service_tier=service_tier,
                    stop=stop,
                    store=store,
                    top_logprobs=top_logprobs,
                    user=user,
                    extra_headers=extra_headers,
                    extra_query=extra_query,
                    extra_body=extra_body,
                    timeout=timeout,
                )

        raw_client = self.raw_client  # 调用完 self.raw_client 后，self.model_id 会被赋值
        assert isinstance(raw_client, AsyncAnthropic | AsyncAnthropicVertex | AsyncAnthropicBedrock)

        _tools = omit if isinstance(tools, NotGiven) else tools
        _tool_choice = omit if isinstance(tool_choice, NotGiven) else tool_choice
        _top_p = omit if isinstance(top_p, NotGiven) or top_p is None else top_p
        if isinstance(self.temperature, NotGiven) or self.temperature is None:
            self.temperature = omit
        _thinking = omit if isinstance(thinking, NotGiven) or thinking is None else thinking

        if messages[0].get("role") == "system":
            system_content = messages[0]["content"]
            # Preserve list format for system prompt to support cache_control
            if isinstance(system_content, list):
                system_prompt = system_content  # Keep as list to preserve cache_control
            else:
                system_prompt = system_content  # String format
            messages = messages[1:]
        else:
            system_prompt = ""

        if not skip_cutoff and self.context_length_control == ContextLengthControlType.Latest:
            messages = cutoff_messages(
                messages,
                max_count=self.model_setting.context_length,
                backend=self.BACKEND_NAME,
                model=self.model,
            )

        messages = format_messages_alternate(messages)

        tools_params: list[AnthropicToolParam] | Omit = refactor_tool_use_params(_tools) if _tools else omit
        tool_choice_param = omit
        if _tool_choice:
            tool_choice_param = refactor_tool_choice(_tool_choice)

        if not max_tokens:
            max_output_tokens = self.model_setting.max_output_tokens
            native_multimodal = self.model_setting.native_multimodal
            token_counts = get_message_token_counts(messages=messages, tools=tools, model=self.model, native_multimodal=native_multimodal)
            if max_output_tokens is not None:
                max_tokens = self.model_setting.context_length - token_counts
                max_tokens = min(max(max_tokens, 1), max_output_tokens)
            else:
                max_tokens = self.model_setting.context_length - token_counts

        await self._acquire_rate_limit(self.endpoint, self.model, messages)

        if self.stream:
            try:
                stream_response = await raw_client.messages.create(
                    model=self.model_id,
                    messages=messages,
                    system=system_prompt,
                    stream=True,
                    temperature=self.temperature,
                    max_tokens=max_tokens,
                    tools=tools_params,
                    tool_choice=tool_choice_param,
                    top_p=_top_p,
                    thinking=_thinking,
                )
            except AnthropicAPIStatusError as e:
                raise APIStatusError(message=e.message, response=e.response, body=e.body) from e
            except AnthropicAPIConnectionError as e:
                raise APIConnectionError(message=e.message, request=e.request) from e

            async def generator():
                stream_state = init_anthropic_stream_state()
                async for chunk in stream_response:
                    message = adapt_anthropic_stream_event(chunk, stream_state)
                    if message is not None:
                        yield message

            return generator()
        else:
            try:
                response = await raw_client.messages.create(
                    model=self.model_id,
                    messages=messages,
                    system=system_prompt,
                    stream=False,
                    temperature=self.temperature,
                    max_tokens=max_tokens,
                    tools=tools_params,
                    tool_choice=tool_choice_param,
                    top_p=_top_p,
                    thinking=_thinking,
                )
            except AnthropicAPIStatusError as e:
                raise APIStatusError(message=e.message, response=e.response, body=e.body) from e
            except AnthropicAPIConnectionError as e:
                raise APIConnectionError(message=e.message, request=e.request) from e

            return build_anthropic_completion_message(list(response.content), response.usage)



from .common import BaseRetrievalClient, BaseAsyncRetrievalClient, build_url, render_template, extract_json_path

__all__ = [
    "BaseRetrievalClient",
    "BaseAsyncRetrievalClient",
    "build_url",
    "render_template",
    "extract_json_path",
]



from .defaults import (
    CONTEXT_LENGTH_CONTROL,
    ENDPOINT_CONCURRENT_REQUESTS,
    ENDPOINT_RPM,
    ENDPOINT_TPM,
    MOONSHOT_DEFAULT_MODEL,
    MOONSHOT_MODELS,
    DEEPSEEK_DEFAULT_MODEL,
    DEEPSEEK_MODELS,
    BAICHUAN_DEFAULT_MODEL,
    BAICHUAN_MODELS,
    GROQ_DEFAULT_MODEL,
    GROQ_MODELS,
    QWEN_DEFAULT_MODEL,
    QWEN_MODELS,
    YI_DEFAULT_MODEL,
    YI_MODELS,
    ZHIPUAI_DEFAULT_MODEL,
    ZHIPUAI_MODELS,
    MISTRAL_DEFAULT_MODEL,
    MISTRAL_MODELS,
    OPENAI_DEFAULT_MODEL,
    OPENAI_MODELS,
    ANTHROPIC_DEFAULT_MODEL,
    ANTHROPIC_MODELS,
    MINIMAX_DEFAULT_MODEL,
    MINIMAX_MODELS,
    GEMINI_DEFAULT_MODEL,
    GEMINI_MODELS,
    ERNIE_DEFAULT_MODEL,
    ERNIE_MODELS,
    STEPFUN_DEFAULT_MODEL,
    STEPFUN_MODELS,
    XAI_DEFAULT_MODEL,
    XAI_MODELS,
)
from .enums import BackendType, EmbeddingBackendType, RerankBackendType, LLMType, ContextLengthControlType
from .exception import APIStatusError, APIConnectionError
from .llm_parameters import (
    EndpointSetting,
    ModelSetting,
    BackendSettings,
    RequestMapping,
    ResponseMapping,
    RetrievalModelSetting,
    RetrievalBackendSettings,
    Usage,
    ChatCompletionMessage,
    ChatCompletionDeltaMessage,
    ChatCompletionStreamOptionsParam,
    NotGiven,
    NOT_GIVEN,
    OpenAIToolParam,
    ToolParam,
    Tools,
    ToolChoice,
    AnthropicToolParam,
    AnthropicToolChoice,
    OPENAI_NOT_GIVEN,
    ANTHROPIC_NOT_GIVEN,
    VectorVeinMessage,
    VectorVeinTextMessage,
    VectorVeinWorkflowMessage,
)
from .retrieval_parameters import (
    EmbeddingData,
    EmbeddingUsage,
    EmbeddingResponse,
    RerankResult,
    RerankUsage,
    RerankResponse,
)
from .settings import (
    RedisConfigDict,
    DiskCacheConfigDict,
    RateLimitConfigDict,
    ServerDict,
    EndpointOptionDict,
    RequestMappingDict,
    ResponseMappingDict,
    ModelConfigDict,
    RetrievalModelConfigDict,
    BackendSettingsDict,
    RetrievalBackendSettingsDict,
    EndpointSettingDict,
    BackendsDict,
    SettingsDict,
    SettingsV1Dict,
    SettingsV2Dict,
)

from anthropic.types import ThinkingConfigParam, ThinkingConfigEnabledParam
from openai.types.chat.completion_create_params import ResponseFormat


__all__ = [
    "CONTEXT_LENGTH_CONTROL",
    "ENDPOINT_CONCURRENT_REQUESTS",
    "ENDPOINT_RPM",
    "ENDPOINT_TPM",
    "MOONSHOT_DEFAULT_MODEL",
    "MOONSHOT_MODELS",
    "DEEPSEEK_DEFAULT_MODEL",
    "DEEPSEEK_MODELS",
    "BAICHUAN_DEFAULT_MODEL",
    "BAICHUAN_MODELS",
    "GROQ_DEFAULT_MODEL",
    "GROQ_MODELS",
    "QWEN_DEFAULT_MODEL",
    "QWEN_MODELS",
    "YI_DEFAULT_MODEL",
    "YI_MODELS",
    "ZHIPUAI_DEFAULT_MODEL",
    "ZHIPUAI_MODELS",
    "MISTRAL_DEFAULT_MODEL",
    "MISTRAL_MODELS",
    "OPENAI_DEFAULT_MODEL",
    "OPENAI_MODELS",
    "ANTHROPIC_DEFAULT_MODEL",
    "ANTHROPIC_MODELS",
    "MINIMAX_DEFAULT_MODEL",
    "MINIMAX_MODELS",
    "GEMINI_DEFAULT_MODEL",
    "GEMINI_MODELS",
    "ERNIE_DEFAULT_MODEL",
    "ERNIE_MODELS",
    "STEPFUN_DEFAULT_MODEL",
    "STEPFUN_MODELS",
    "XAI_DEFAULT_MODEL",
    "XAI_MODELS",
    "BackendType",
    "EmbeddingBackendType",
    "RerankBackendType",
    "LLMType",
    "ContextLengthControlType",
    "APIStatusError",
    "APIConnectionError",
    "EndpointOptionDict",
    "EndpointSetting",
    "ModelSetting",
    "BackendSettings",
    "RequestMapping",
    "ResponseMapping",
    "RetrievalModelSetting",
    "RetrievalBackendSettings",
    "Usage",
    "EmbeddingData",
    "EmbeddingUsage",
    "EmbeddingResponse",
    "RerankResult",
    "RerankUsage",
    "RerankResponse",
    "ChatCompletionMessage",
    "ChatCompletionDeltaMessage",
    "ChatCompletionStreamOptionsParam",
    "AnthropicToolParam",
    "AnthropicToolChoice",
    "OPENAI_NOT_GIVEN",
    "ANTHROPIC_NOT_GIVEN",
    "VectorVeinMessage",
    "VectorVeinTextMessage",
    "VectorVeinWorkflowMessage",
    "NotGiven",
    "NOT_GIVEN",
    "OpenAIToolParam",
    "ToolParam",
    "Tools",
    "ToolChoice",
    "RedisConfigDict",
    "DiskCacheConfigDict",
    "RateLimitConfigDict",
    "ServerDict",
    "EndpointOptionDict",
    "RequestMappingDict",
    "ResponseMappingDict",
    "ModelConfigDict",
    "RetrievalModelConfigDict",
    "BackendSettingsDict",
    "RetrievalBackendSettingsDict",
    "EndpointSettingDict",
    "BackendsDict",
    "SettingsDict",
    "SettingsV1Dict",
    "SettingsV2Dict",
    "ThinkingConfigParam",
    "ThinkingConfigEnabledParam",
    "ResponseFormat",
]



from __future__ import annotations

import asyncio
import random
import re
import time
from collections import defaultdict
from typing import Any, cast

import httpx

from ..settings import Settings, normalize_settings
from ..types.llm_parameters import EndpointSetting, RetrievalBackendSettings
from ..types.settings import EndpointOptionDict, SettingsDict
from ..utilities.rate_limiter import AsyncDiskCacheRateLimiter, AsyncMemoryRateLimiter, AsyncRedisRateLimiter
from ..utilities.rate_limiter import SyncDiskCacheRateLimiter, SyncMemoryRateLimiter, SyncRedisRateLimiter

_PLACEHOLDER_RE = re.compile(r"\$\{([a-zA-Z_][a-zA-Z0-9_]*)\}")


class BaseRetrievalClient:
    def __init__(
        self,
        *,
        model: str,
        backend_name: str,
        backend_settings: RetrievalBackendSettings,
        random_endpoint: bool = True,
        endpoint_id: str = "",
        http_client: httpx.Client | None = None,
        settings: Settings | SettingsDict | None = None,
    ):
        self.model = model or self._resolve_default_model(backend_settings)
        self.backend_name = backend_name
        self.backend_settings = backend_settings
        self.random_endpoint = random_endpoint
        self.endpoint_id = endpoint_id
        self.http_client = http_client

        self.settings: Settings = normalize_settings(settings)

        self.rate_limiter = self._init_rate_limiter()
        self.active_requests = defaultdict(int)
        self.rpm = None
        self.tpm = None
        self.concurrent_requests = None

        self.model_setting = self.backend_settings.get_model_setting(self.model)
        self.model_id = self.model_setting.id
        self.endpoint: EndpointSetting | None = None

        if endpoint_id:
            self.random_endpoint = False
            self.endpoint = self.settings.get_endpoint(self.endpoint_id)
            self._set_model_id_by_endpoint_id(self.endpoint_id)

    @staticmethod
    def _resolve_default_model(backend_settings: RetrievalBackendSettings) -> str:
        for model_name, model_setting in backend_settings.models.items():
            if model_setting.enabled:
                return model_name
        if backend_settings.models:
            return next(iter(backend_settings.models.keys()))
        raise ValueError("No models configured for retrieval backend.")

    def _init_rate_limiter(self):
        if not self.settings.rate_limit:
            return None
        if not self.settings.rate_limit.enabled:
            return None

        if self.settings.rate_limit.backend == "memory":
            return SyncMemoryRateLimiter()
        if self.settings.rate_limit.backend == "redis":
            if not self.settings.rate_limit.redis:
                raise ValueError("Redis settings must be provided if Redis backend is selected.")
            return SyncRedisRateLimiter(
                host=self.settings.rate_limit.redis.host,
                port=self.settings.rate_limit.redis.port,
                db=self.settings.rate_limit.redis.db,
            )
        if self.settings.rate_limit.backend == "diskcache":
            if not self.settings.rate_limit.diskcache:
                raise ValueError("Diskcache settings must be provided if Diskcache backend is selected.")
            return SyncDiskCacheRateLimiter(
                cache_dir=self.settings.rate_limit.diskcache.cache_dir,
            )
        return None

    @staticmethod
    def _estimate_tokens(payload: Any) -> int:
        text = str(payload)
        cjk_chars = sum(1 for ch in text if "\u4e00" <= ch <= "\u9fff" or "\u3000" <= ch <= "\u303f" or "\uff00" <= ch <= "\uffef")
        ascii_chars = len(text) - cjk_chars
        return max(int(cjk_chars * 1.5 + ascii_chars * 0.25), 1)

    def _acquire_rate_limit(self, endpoint: EndpointSetting | None, payload: Any):
        if endpoint is None:
            return

        key = f"{endpoint.id}:{self.model}"
        rpm = self.rpm or endpoint.rpm or (self.settings.rate_limit.default_rpm if self.settings.rate_limit else 60)
        tpm = self.tpm or endpoint.tpm or (self.settings.rate_limit.default_tpm if self.settings.rate_limit else 1000000)
        est_tokens = self._estimate_tokens(payload)

        while self.rate_limiter:
            allowed, wait_time = self.rate_limiter.check_limit(key, rpm, tpm, est_tokens)
            if allowed:
                break
            time.sleep(wait_time)

    def _get_available_endpoints(self, model_endpoints: list[str | EndpointOptionDict]) -> list[str | EndpointOptionDict]:
        available_endpoints = []
        for endpoint_option in model_endpoints:
            if isinstance(endpoint_option, str):
                endpoint_id = endpoint_option
            else:
                endpoint_id = endpoint_option.get("endpoint_id")
                if not isinstance(endpoint_id, str):
                    continue
            try:
                endpoint = self.settings.get_endpoint(endpoint_id)
            except ValueError:
                continue
            if endpoint.enabled:
                available_endpoints.append(endpoint_option)
        return available_endpoints

    def _set_model_id_by_endpoint_id(self, endpoint_id: str) -> str:
        self.model_id = self.model_setting.id
        for endpoint_option in self.model_setting.endpoints:
            if isinstance(endpoint_option, dict) and endpoint_option.get("endpoint_id") == endpoint_id:
                self.model_id = endpoint_option.get("model_id", self.model_setting.id)
                break
        return self.model_id

    def _set_endpoint(self) -> tuple[EndpointSetting, str]:
        if self.endpoint is None:
            if self.random_endpoint:
                available_endpoints = self._get_available_endpoints(
                    cast(list[str | EndpointOptionDict], self.model_setting.endpoints)
                )
                if not available_endpoints:
                    raise ValueError(f"No enabled endpoints available for model {self.model}")

                endpoint_option = random.choice(available_endpoints)
                if isinstance(endpoint_option, dict):
                    self.endpoint_id = endpoint_option["endpoint_id"]
                    self.model_id = endpoint_option.get("model_id", self.model_setting.id)
                    self.rpm = endpoint_option.get("rpm", None)
                    self.tpm = endpoint_option.get("tpm", None)
                    self.concurrent_requests = endpoint_option.get("concurrent_requests", None)
                else:
                    self.endpoint_id = endpoint_option
                    self.model_id = self.model_setting.id
                self.endpoint = self.settings.get_endpoint(self.endpoint_id)
            else:
                if not self.endpoint_id:
                    raise ValueError("endpoint_id is required when random_endpoint is False")
                self.endpoint = self.settings.get_endpoint(self.endpoint_id)
                if not self.endpoint.enabled:
                    raise ValueError(f"Endpoint {self.endpoint_id} is disabled")
                self._set_model_id_by_endpoint_id(self.endpoint_id)
        else:
            if not self.endpoint.enabled:
                raise ValueError(f"Endpoint {self.endpoint.id} is disabled")
            self.endpoint_id = self.endpoint.id
            self._set_model_id_by_endpoint_id(self.endpoint_id)

        return self.endpoint, self.model_id


class BaseAsyncRetrievalClient:
    def __init__(
        self,
        *,
        model: str,
        backend_name: str,
        backend_settings: RetrievalBackendSettings,
        random_endpoint: bool = True,
        endpoint_id: str = "",
        http_client: httpx.AsyncClient | None = None,
        settings: Settings | SettingsDict | None = None,
    ):
        self.model = model or self._resolve_default_model(backend_settings)
        self.backend_name = backend_name
        self.backend_settings = backend_settings
        self.random_endpoint = random_endpoint
        self.endpoint_id = endpoint_id
        self.http_client = http_client

        self.settings: Settings = normalize_settings(settings)

        self.rate_limiter = self._init_rate_limiter()
        self.active_requests = defaultdict(int)
        self.rpm = None
        self.tpm = None
        self.concurrent_requests = None

        self.model_setting = self.backend_settings.get_model_setting(self.model)
        self.model_id = self.model_setting.id
        self.endpoint: EndpointSetting | None = None

        if endpoint_id:
            self.random_endpoint = False
            self.endpoint = self.settings.get_endpoint(self.endpoint_id)
            self._set_model_id_by_endpoint_id(self.endpoint_id)

    @staticmethod
    def _resolve_default_model(backend_settings: RetrievalBackendSettings) -> str:
        for model_name, model_setting in backend_settings.models.items():
            if model_setting.enabled:
                return model_name
        if backend_settings.models:
            return next(iter(backend_settings.models.keys()))
        raise ValueError("No models configured for retrieval backend.")

    def _init_rate_limiter(self):
        if not self.settings.rate_limit:
            return None
        if not self.settings.rate_limit.enabled:
            return None

        if self.settings.rate_limit.backend == "memory":
            return AsyncMemoryRateLimiter()
        if self.settings.rate_limit.backend == "redis":
            if not self.settings.rate_limit.redis:
                raise ValueError("Redis settings must be provided if Redis backend is selected.")
            return AsyncRedisRateLimiter(
                host=self.settings.rate_limit.redis.host,
                port=self.settings.rate_limit.redis.port,
                db=self.settings.rate_limit.redis.db,
            )
        if self.settings.rate_limit.backend == "diskcache":
            if not self.settings.rate_limit.diskcache:
                raise ValueError("Diskcache settings must be provided if Diskcache backend is selected.")
            return AsyncDiskCacheRateLimiter(
                cache_dir=self.settings.rate_limit.diskcache.cache_dir,
            )
        return None

    @staticmethod
    def _estimate_tokens(payload: Any) -> int:
        text = str(payload)
        cjk_chars = sum(1 for ch in text if "\u4e00" <= ch <= "\u9fff" or "\u3000" <= ch <= "\u303f" or "\uff00" <= ch <= "\uffef")
        ascii_chars = len(text) - cjk_chars
        return max(int(cjk_chars * 1.5 + ascii_chars * 0.25), 1)

    async def _acquire_rate_limit(self, endpoint: EndpointSetting | None, payload: Any):
        if endpoint is None:
            return

        key = f"{endpoint.id}:{self.model}"
        rpm = self.rpm or endpoint.rpm or (self.settings.rate_limit.default_rpm if self.settings.rate_limit else 60)
        tpm = self.tpm or endpoint.tpm or (self.settings.rate_limit.default_tpm if self.settings.rate_limit else 1000000)
        est_tokens = self._estimate_tokens(payload)

        while self.rate_limiter:
            allowed, wait_time = await self.rate_limiter.check_limit(key, rpm, tpm, est_tokens)
            if allowed:
                break
            await asyncio.sleep(wait_time)

    def _get_available_endpoints(self, model_endpoints: list[str | EndpointOptionDict]) -> list[str | EndpointOptionDict]:
        available_endpoints = []
        for endpoint_option in model_endpoints:
            if isinstance(endpoint_option, str):
                endpoint_id = endpoint_option
            else:
                endpoint_id = endpoint_option.get("endpoint_id")
                if not isinstance(endpoint_id, str):
                    continue
            try:
                endpoint = self.settings.get_endpoint(endpoint_id)
            except ValueError:
                continue
            if endpoint.enabled:
                available_endpoints.append(endpoint_option)
        return available_endpoints

    def _set_model_id_by_endpoint_id(self, endpoint_id: str) -> str:
        self.model_id = self.model_setting.id
        for endpoint_option in self.model_setting.endpoints:
            if isinstance(endpoint_option, dict) and endpoint_option.get("endpoint_id") == endpoint_id:
                self.model_id = endpoint_option.get("model_id", self.model_setting.id)
                break
        return self.model_id

    def _set_endpoint(self) -> tuple[EndpointSetting, str]:
        if self.endpoint is None:
            if self.random_endpoint:
                available_endpoints = self._get_available_endpoints(
                    cast(list[str | EndpointOptionDict], self.model_setting.endpoints)
                )
                if not available_endpoints:
                    raise ValueError(f"No enabled endpoints available for model {self.model}")

                endpoint_option = random.choice(available_endpoints)
                if isinstance(endpoint_option, dict):
                    self.endpoint_id = endpoint_option["endpoint_id"]
                    self.model_id = endpoint_option.get("model_id", self.model_setting.id)
                    self.rpm = endpoint_option.get("rpm", None)
                    self.tpm = endpoint_option.get("tpm", None)
                    self.concurrent_requests = endpoint_option.get("concurrent_requests", None)
                else:
                    self.endpoint_id = endpoint_option
                    self.model_id = self.model_setting.id
                self.endpoint = self.settings.get_endpoint(self.endpoint_id)
            else:
                if not self.endpoint_id:
                    raise ValueError("endpoint_id is required when random_endpoint is False")
                self.endpoint = self.settings.get_endpoint(self.endpoint_id)
                if not self.endpoint.enabled:
                    raise ValueError(f"Endpoint {self.endpoint_id} is disabled")
                self._set_model_id_by_endpoint_id(self.endpoint_id)
        else:
            if not self.endpoint.enabled:
                raise ValueError(f"Endpoint {self.endpoint.id} is disabled")
            self.endpoint_id = self.endpoint.id
            self._set_model_id_by_endpoint_id(self.endpoint_id)

        return self.endpoint, self.model_id


def build_url(api_base: str | None, path: str) -> str:
    if path.startswith("http://") or path.startswith("https://"):
        return path
    if api_base is None:
        raise ValueError("api_base is required for relative request paths.")
    return f"{api_base.rstrip('/')}/{path.lstrip('/')}"


def render_template(value: Any, context: dict[str, Any]) -> Any:
    if isinstance(value, dict):
        return {k: render_template(v, context) for k, v in value.items()}
    if isinstance(value, list):
        return [render_template(v, context) for v in value]
    if isinstance(value, str):
        matches = list(_PLACEHOLDER_RE.finditer(value))
        if not matches:
            return value
        if len(matches) == 1 and matches[0].span() == (0, len(value)):
            return context.get(matches[0].group(1))

        result = value
        for match in matches:
            key = match.group(1)
            result = result.replace(f"${{{key}}}", str(context.get(key, "")))
        return result
    return value


def _tokenize_path(path: str) -> list[tuple[str, str | int]]:
    if path == "$":
        return []
    if not path.startswith("$"):
        raise ValueError(f"Path must start with '$': {path}")

    tokens: list[tuple[str, str | int]] = []
    i = 1
    n = len(path)

    while i < n:
        ch = path[i]
        if ch == ".":
            i += 1
            start = i
            while i < n and path[i] not in ".[":
                i += 1
            key = path[start:i]
            if key:
                tokens.append(("key", key))
            continue

        if ch == "[":
            j = path.find("]", i)
            if j == -1:
                raise ValueError(f"Invalid path: {path}")
            content = path[i + 1 : j]
            if content == "*":
                tokens.append(("all", "*"))
            elif content.isdigit():
                tokens.append(("index", int(content)))
            elif (content.startswith("'") and content.endswith("'")) or (content.startswith('"') and content.endswith('"')):
                tokens.append(("key", content[1:-1]))
            else:
                raise ValueError(f"Unsupported bracket selector in path: {path}")
            i = j + 1
            continue

        i += 1

    return tokens


def extract_json_path(data: Any, path: str | None) -> Any:
    if path is None:
        return None

    tokens = _tokenize_path(path)
    values = [data]
    has_all = any(kind == "all" for kind, _ in tokens)

    for kind, token in tokens:
        next_values: list[Any] = []
        for current in values:
            if kind == "key":
                if isinstance(current, dict) and token in current:
                    next_values.append(current[token])
            elif kind == "index":
                if isinstance(current, list) and isinstance(token, int) and 0 <= token < len(current):
                    next_values.append(current[token])
            elif kind == "all":
                if isinstance(current, list):
                    next_values.extend(current)
        values = next_values

    if not values:
        return None
    if has_all:
        return values
    if len(values) == 1:
        return values[0]
    return values



# @Author: Bi Ying
# @Date:   2024-07-26 14:48:55
import re
import json
import uuid
import warnings
import logging
from math import ceil
from collections.abc import Iterable
from typing import Any, cast

import httpx
import tiktoken
from anthropic import Anthropic, AnthropicVertex
from anthropic._types import NOT_GIVEN as ANTHROPIC_NOT_GIVEN
from openai.types.chat import ChatCompletionMessageParam

from ..settings import settings
from ..utilities.retry import Retry
from ..utilities.gcp_token import get_token_with_cache
from ..types.enums import BackendType
from ..types.llm_parameters import (
    NotGiven,
    NOT_GIVEN,
    ToolParam,
    VectorVeinMessage,
    VectorVeinWorkflowMessage,
)


gpt_35_encoding = None
gpt_4o_encoding = None
logger = logging.getLogger(__name__)


def get_gpt_35_encoding():
    global gpt_35_encoding
    if gpt_35_encoding is None:
        gpt_35_encoding = tiktoken.encoding_for_model("gpt-3.5-turbo")
    return gpt_35_encoding


def get_gpt_4o_encoding():
    global gpt_4o_encoding
    if gpt_4o_encoding is None:
        gpt_4o_encoding = tiktoken.encoding_for_model("gpt-4o")
    return gpt_4o_encoding


class ToolCallContentProcessor:
    tool_use_re = re.compile(r"<\|▶\|>(.*?)<\|◀\|>", re.DOTALL)

    def __init__(self, content: str):
        self.content = content

    @property
    def non_tool_content(self):
        return re.sub(self.tool_use_re, "", self.content).strip()

    @property
    def tool_calls(self):
        if "<|▶|>" not in self.content or "<|◀|>" not in self.content:
            return {}
        tool_calls_matches = re.findall(self.tool_use_re, self.content)
        if tool_calls_matches:
            tool_calls = []
            for idx, match in enumerate(tool_calls_matches):
                try:
                    tool_call_data = json.loads(match)
                    arguments = json.dumps(tool_call_data["arguments"], ensure_ascii=False)
                    tool_calls.append(
                        {
                            "index": idx,
                            "id": f"tool_call_{uuid.uuid4()}",
                            "function": {
                                "arguments": arguments,
                                "name": tool_call_data["name"],
                            },
                            "type": "function",
                        }
                    )
                except json.JSONDecodeError:
                    logger.warning("Failed to parse tool call data. content=%s match=%s", self.content, match)

            if not tool_calls:
                return {}

            return {"tool_calls": tool_calls}
        else:
            return {}


def convert_type(value, value_type):
    if value_type == "string":
        return str(value)
    elif value_type == "number":
        try:
            return float(value)
        except ValueError:
            return value
    elif value_type == "integer":
        try:
            return int(value)
        except ValueError:
            return value
    elif value_type == "boolean":
        return value.lower() in ("true", "1", "t")
    else:
        return value  # 如果类型未知，返回原始值


def _get_first_enabled_endpoint(backend_setting, settings):
    """Get the first enabled endpoint from backend settings"""
    for endpoint_choice in backend_setting.endpoints:
        if isinstance(endpoint_choice, dict):
            endpoint_id = endpoint_choice["endpoint_id"]
        else:
            endpoint_id = endpoint_choice

        try:
            endpoint = settings.get_endpoint(endpoint_id)
            if endpoint.enabled:
                return endpoint
        except ValueError:
            # Endpoint not found, skip it
            continue
    return None


def get_token_counts(text: str | dict, model: str = "", use_token_server_first: bool = True) -> int:
    if use_token_server_first and settings.token_server is not None:
        base_url = settings.token_server.url if settings.token_server.url is not None else f"http://{settings.token_server.host}:{settings.token_server.port}"
        _, response = Retry(httpx.post).args(url=f"{base_url}/count_tokens", json={"text": text, "model": model}, timeout=None).retry_times(5).sleep_time(1).run()
        if response is not None:
            try:
                result = response.json()
                return result["total_tokens"]
            except Exception as e:
                logger.warning("Failed to parse token count response: %s", e)

    if not isinstance(text, str):
        text = str(text)
    if model == "gpt-3.5-turbo":
        return len(get_gpt_35_encoding().encode(text))
    elif model.startswith(("gpt-4o", "o1-", "o3-")):
        return len(get_gpt_4o_encoding().encode(text))
    elif model.startswith(("abab", "MiniMax")):
        backend_setting = settings.get_backend(BackendType.MiniMax).models[model]
        if len(backend_setting.endpoints) == 0:
            return int(len(text) / 1.33)
        endpoint = _get_first_enabled_endpoint(backend_setting, settings)
        if endpoint is None:
            return int(len(text) / 1.33)
        tokenize_url = "https://api.minimax.chat/v1/tokenize"
        headers = {"Authorization": f"Bearer {endpoint.api_key}", "Content-Type": "application/json"}
        request_body = {
            "model": model,
            "tokens_to_generate": 128,
            "temperature": 0.2,
            "messages": [
                {"sender_type": "USER", "text": text},
            ],
        }

        _, response = (
            Retry(httpx.post)
            .args(
                url=tokenize_url,
                headers=headers,
                json=request_body,
                timeout=None,
                proxy=endpoint.proxy or None,
            )
            .retry_times(5)
            .sleep_time(10)
            .run()
        )
        if response is None:
            return 1000
        result = response.json()
        return result["segments_num"]
    elif model.startswith(("moonshot", "kimi")):
        backend_setting = settings.get_backend(BackendType.Moonshot).models[model]
        if len(backend_setting.endpoints) == 0:
            return len(get_gpt_35_encoding().encode(text))
        endpoint = _get_first_enabled_endpoint(backend_setting, settings)
        if endpoint is None:
            return len(get_gpt_35_encoding().encode(text))
        tokenize_url = f"{endpoint.api_base}/tokenizers/estimate-token-count"
        headers = {"Content-Type": "application/json", "Authorization": f"Bearer {endpoint.api_key}"}
        request_body = {
            "model": model,
            "messages": [
                {"role": "user", "content": text},
            ],
        }
        _, response = (
            Retry(httpx.post)
            .args(
                url=tokenize_url,
                headers=headers,
                json=request_body,
                timeout=None,
                proxy=endpoint.proxy or None,
            )
            .retry_times(5)
            .sleep_time(10)
            .run()
        )
        if response is None:
            return 1000
        result = response.json()
        return result["data"]["total_tokens"]
    elif model.startswith("gemini"):
        backend_setting = settings.get_backend(BackendType.Gemini).models[model]
        if len(backend_setting.endpoints) == 0:
            return len(get_gpt_35_encoding().encode(text))
        endpoint = _get_first_enabled_endpoint(backend_setting, settings)
        if endpoint is None:
            return len(get_gpt_35_encoding().encode(text))

        api_base = endpoint.api_base.removesuffix("/openai/") if endpoint.api_base else "https://generativelanguage.googleapis.com/v1beta"
        if backend_setting.id.startswith("gemini-3"):
            base_url = f"{api_base}/models/gemini-2.5-pro:countTokens"
        else:
            base_url = f"{api_base}/models/{backend_setting.id}:countTokens"
        params = {"key": endpoint.api_key}
        request_body = {
            "contents": {
                "role": "USER",
                "parts": [
                    {"text": text},
                ],
            },
        }
        _, response = (
            Retry(httpx.post)
            .args(
                base_url,
                json=request_body,
                params=params,
                timeout=None,
                proxy=endpoint.proxy or None,
            )
            .retry_times(5)
            .sleep_time(10)
            .run()
        )
        if response is None:
            return 1000
        result = response.json()
        return result["totalTokens"]
    elif model.startswith("claude"):
        backend_setting = settings.get_backend(BackendType.Anthropic)
        try:
            for endpoint_choice in backend_setting.models[model].endpoints:
                if isinstance(endpoint_choice, dict):
                    endpoint_id = endpoint_choice["endpoint_id"]
                else:
                    endpoint_id = endpoint_choice
                endpoint_id = cast(str, endpoint_id)

                try:
                    endpoint = settings.get_endpoint(endpoint_id)
                    if not endpoint.enabled:
                        continue
                except ValueError:
                    continue

                http_client = httpx.Client(proxy=endpoint.proxy) if endpoint.proxy else None

                if endpoint.is_bedrock or endpoint.endpoint_type == "anthropic_bedrock":
                    try:
                        import boto3

                        if endpoint.credentials is None:
                            warnings.warn("Anthropic Bedrock endpoint requires credentials", stacklevel=2)
                            continue
                        bedrock_credentials = cast(dict[str, Any], endpoint.credentials)
                        client = boto3.client(
                            "bedrock-runtime",
                            region_name=endpoint.region,
                            aws_access_key_id=bedrock_credentials.get("access_key"),
                            aws_secret_access_key=bedrock_credentials.get("secret_key"),
                        )
                        tokens = client.count_tokens(
                            # modelId=endpoint_choice["model_id"],
                            modelId="anthropic.claude-sonnet-4-20250514-v1:0",
                            input={
                                "converse": {
                                    "messages": [
                                        {
                                            "role": "user",
                                            "content": [{"text": text}],
                                        }
                                    ]
                                },
                            },
                        ).get("inputTokens")
                        if tokens is None:
                            continue
                        else:
                            return tokens
                    except Exception as e:
                        warnings.warn(f"Anthropic Bedrock token counting initialization failed: {e}", stacklevel=2)
                        continue
                elif endpoint.is_vertex or endpoint.endpoint_type == "anthropic_vertex":
                    if endpoint.credentials is None:
                        warnings.warn("Anthropic Vertex endpoint requires credentials", stacklevel=2)
                        continue

                    access_token, expires_at = get_token_with_cache(
                        credentials=endpoint.credentials,
                        proxy=endpoint.proxy,
                        cached_token=endpoint.access_token,
                        cached_expires_at=endpoint.access_token_expires_at,
                    )
                    endpoint.access_token = access_token
                    endpoint.access_token_expires_at = expires_at

                    if endpoint.api_base is None:
                        base_url = None
                    else:
                        base_url = f"{endpoint.api_base}{endpoint.region}-aiplatform/v1"

                    region = ANTHROPIC_NOT_GIVEN if endpoint.region is None else endpoint.region
                    return (
                        AnthropicVertex(
                            region=region,
                            base_url=base_url,
                            project_id=endpoint.credentials.get("quota_project_id", NOT_GIVEN),
                            access_token=access_token,
                            http_client=http_client,
                        )
                        .messages.count_tokens(messages=[{"role": "user", "content": text}], model=model)
                        .input_tokens
                    )
                elif endpoint.endpoint_type in ("default", "anthropic"):
                    return (
                        Anthropic(
                            api_key=endpoint.api_key,
                            base_url=endpoint.api_base,
                            http_client=http_client,
                        )
                        .messages.count_tokens(messages=[{"role": "user", "content": text}], model=model)
                        .input_tokens
                    )
        except Exception as e:
            warnings.warn(f"Anthropic token counting failed: {e}", stacklevel=2)

        # TODO: Use anthropic token counting
        warnings.warn("Fallback to GPT-4o token counting", stacklevel=2)
        return len(get_gpt_4o_encoding().encode(text))
    elif model.startswith("deepseek"):
        from deepseek_tokenizer import deepseek_tokenizer

        return len(deepseek_tokenizer.encode(text))
    elif model.startswith("qwen"):
        from qwen_tokenizer import get_tokenizer

        qwen_tokenizer = get_tokenizer(model)
        return len(qwen_tokenizer.encode(text))
    elif model.startswith("stepfun"):
        backend_setting = settings.get_backend(BackendType.StepFun).models[model]
        if len(backend_setting.endpoints) == 0:
            return len(get_gpt_35_encoding().encode(text))
        endpoint = _get_first_enabled_endpoint(backend_setting, settings)
        if endpoint is None:
            return len(get_gpt_35_encoding().encode(text))
        tokenize_url = f"{endpoint.api_base}/token/count"
        headers = {"Content-Type": "application/json", "Authorization": f"Bearer {endpoint.api_key}"}
        request_body = {
            "model": model,
            "messages": [
                {"role": "user", "content": text},
            ],
        }
        _, response = (
            Retry(httpx.post)
            .args(
                url=tokenize_url,
                headers=headers,
                json=request_body,
                timeout=None,
                proxy=endpoint.proxy or None,
            )
            .retry_times(5)
            .sleep_time(10)
            .run()
        )
        if response is None:
            return 1000
        result = response.json()
        return result["data"]["total_tokens"]
    elif model.startswith("glm"):
        backend_setting = settings.get_backend(BackendType.ZhiPuAI).models[model]
        if len(backend_setting.endpoints) == 0:
            return len(get_gpt_35_encoding().encode(text))
        endpoint = _get_first_enabled_endpoint(backend_setting, settings)
        if endpoint is None:
            return len(get_gpt_35_encoding().encode(text))
        if model not in ("glm-4-plus", "glm-4-long", "glm-4-0520", "glm-4-air", "glm-4-flash"):
            model = "glm-4-plus"
        tokenize_url = f"{endpoint.api_base or 'https://open.bigmodel.cn/api/paas/v4'}/tokenizer"
        headers = {"Content-Type": "application/json", "Authorization": f"Bearer {endpoint.api_key}"}
        request_body = {
            "model": model,
            "messages": [
                {"role": "user", "content": text},
            ],
        }
        _, response = (
            Retry(httpx.post)
            .args(
                url=tokenize_url,
                headers=headers,
                json=request_body,
                timeout=None,
                proxy=endpoint.proxy or None,
            )
            .result_check(lambda response: "usage" in response.json())
            .retry_times(5)
            .sleep_time(10)
            .run()
        )
        if response is None:
            return 1000
        result = response.json()
        return result["usage"]["prompt_tokens"]
    else:
        return len(get_gpt_4o_encoding().encode(text))


def calculate_image_tokens(width: int, height: int, model: str = "gpt-4o"):
    if model.startswith("moonshot"):
        return 1024

    if width > 2048 or height > 2048:
        aspect_ratio = width / height
        if aspect_ratio > 1:
            width, height = 2048, int(2048 / aspect_ratio)
        else:
            width, height = int(2048 * aspect_ratio), 2048

    if width >= height and height > 768:
        width, height = int((768 / height) * width), 768
    elif height > width and width > 768:
        width, height = 768, int((768 / width) * height)

    tiles_width = ceil(width / 512)
    tiles_height = ceil(height / 512)
    total_tokens = 85 + 170 * (tiles_width * tiles_height)

    return total_tokens


def get_message_token_counts(
    messages: list,
    tools: list | Iterable[ToolParam] | NotGiven = NOT_GIVEN,
    model: str = "gpt-4o",
    native_multimodal: bool = True,
) -> int:
    formatted_messages = format_messages(messages, backend=BackendType.OpenAI, native_multimodal=native_multimodal)

    # 合并所有文本内容
    all_text_content = []
    image_count = 0

    for message in formatted_messages:
        content = message["content"]
        if isinstance(content, str):
            all_text_content.append(content)
        elif isinstance(content, list):
            for item in content:
                if isinstance(item, dict) and item["type"] == "text":
                    all_text_content.append(item["text"])
                elif isinstance(item, dict) and item["type"].startswith("image"):
                    image_count += 1

    # 一次性计算所有文本的 tokens
    combined_text = "\n".join(all_text_content)
    tokens = get_token_counts(combined_text, model) if combined_text else 0

    # 一次性计算所有图片的 tokens
    if image_count > 0:
        if native_multimodal:
            # TODO: Get real image size
            tokens += image_count * calculate_image_tokens(2048, 2048, model)
        else:
            tokens += image_count

    # 计算 tools tokens
    if tools:
        tokens += get_token_counts(str(tools), model)

    return tokens


def cutoff_messages(
    messages: list,
    max_count: int = 16000,
    backend: BackendType = BackendType.OpenAI,
    model: str = "",
) -> list:
    """
    给定一个消息列表和最大长度，将消息列表截断到最大长度。
    如果列表中第一个元素的role是'system'，则始终保留该元素。
    超过长度时从列表开始处（第二个元素起）依次删除消息，直到总长度小于等于最大长度。
    如果最后一条消息超过了最大长度，那么将最后一条消息截断到最大长度。

    Args:
        messages (list): 消息列表，每条消息是一个包含'role'和'content'的字典。
        max_count (int, optional): 允许的最大长度。默认值为16000。

    Returns:
        list: 截断后的消息列表。
    """

    if len(messages) == 0:
        return messages

    messages_length = 0
    content_key = "content"

    # 先检查并保留第一条system消息（如果有）
    system_message = None
    if messages[0]["role"] == "system":
        system_message = messages[0]
        system_message_length = get_token_counts(system_message[content_key], model)
        if system_message_length > max_count:
            # 如果第一条system消息超过最大长度，截断它
            system_message[content_key] = system_message[content_key][-max_count:]
            return [system_message]
        else:
            messages_length += system_message_length
            messages = messages[1:]  # 移除第一个元素，以处理其余消息

    if system_message:
        system_message = [system_message]
    else:
        system_message = []

    for index, message in enumerate(reversed(messages)):
        if not message.get(content_key):
            continue
        count = 0
        if isinstance(message[content_key], str):
            contents = [message[content_key]]
        elif isinstance(message[content_key], list):
            contents = message[content_key]
        else:
            contents = [str(message[content_key])]

        for content in contents:
            # TODO: Add non text token counts
            if isinstance(content, dict) and "text" not in content:
                continue
            if isinstance(content, dict):
                content_text = content["text"]
            else:
                content_text = str(content)
            count += get_token_counts(content_text, model)
        messages_length += count
        if messages_length < max_count:
            continue
        if index == 0:
            # 一条消息就超过长度则将该消息内容进行截断，保留该消息最后的一部分
            content = message[content_key][max_count - messages_length :]
            return system_message + [
                {
                    "role": message["role"],
                    content_key: content,
                }
            ]

        return system_message + messages[-index:]
    return system_message + messages


def format_image_message(
    image: str,
    backend: BackendType = BackendType.OpenAI,
    process_image: bool = True,
    cache_control: dict | None = None,
) -> dict:
    if process_image:
        from ..utilities.media_processing import ImageProcessor

        image_processor = ImageProcessor(image_source=image)
        if backend == BackendType.OpenAI:
            result = {
                "type": "image_url",
                "image_url": {"url": image_processor.data_url},
            }
        elif backend == BackendType.Anthropic:
            result = {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": image_processor.mime_type,
                    "data": image_processor.base64_image,
                },
            }
        else:
            result = {
                "type": "image_url",
                "image_url": {"url": image_processor.data_url},
            }
    else:
        if backend == BackendType.Anthropic:
            result = {"type": "image", "source": {"type": "base64", "media_type": "image/jpeg", "data": image}}
        else:
            result = {"type": "image_url", "image_url": {"url": image}}

    # Preserve cache_control if provided
    if cache_control is not None:
        result["cache_control"] = cache_control

    return result


def format_workflow_messages(message: VectorVeinWorkflowMessage, content: str, backend: BackendType):
    formatted_messages = []
    metadata = message.get("metadata", {}) if isinstance(message, dict) else {}
    selected_workflow = metadata.get("selected_workflow", {}) if isinstance(metadata, dict) else {}

    # 尝试保留原始 tool_calls（其中可能包含 Gemini 的 thought_signature）
    preserved_tool_calls: list[dict] = []
    if isinstance(metadata, dict) and isinstance(metadata.get("tool_calls"), list):
        preserved_tool_calls = [tc for tc in metadata["tool_calls"] if isinstance(tc, dict)]
    if not preserved_tool_calls:
        content_block = message.get("content", {}) if isinstance(message, dict) else {}
        if isinstance(content_block, dict) and isinstance(content_block.get("tool_calls"), list):
            preserved_tool_calls = [tc for tc in content_block["tool_calls"] if isinstance(tc, dict)]

    preserved_tool_call = preserved_tool_calls[0] if preserved_tool_calls else None
    preserved_func = preserved_tool_call.get("function") if isinstance(preserved_tool_call, dict) else {}

    tool_call_id = None
    if isinstance(preserved_tool_call, dict) and preserved_tool_call.get("id"):
        tool_call_id = preserved_tool_call["id"]
    if tool_call_id is None and isinstance(selected_workflow, dict):
        tool_call_id = selected_workflow.get("tool_call_id") or metadata.get("record_id")

    tool_name = None
    if isinstance(preserved_func, dict) and preserved_func.get("name"):
        tool_name = preserved_func["name"]
    if tool_name is None and isinstance(selected_workflow, dict):
        tool_name = selected_workflow.get("function_name")

    tool_arguments = None
    if isinstance(preserved_func, dict) and preserved_func.get("arguments"):
        tool_arguments = preserved_func["arguments"]
    if tool_arguments is None and isinstance(selected_workflow, dict):
        tool_arguments = json.dumps(selected_workflow.get("params", {}), ensure_ascii=False)

    extra_content = preserved_tool_call.get("extra_content") if isinstance(preserved_tool_call, dict) else None

    # Extract cache_control from preserved_tool_call if present
    tool_cache_control = preserved_tool_call.get("cache_control") if isinstance(preserved_tool_call, dict) else None

    # 工具调用消息
    if backend in (BackendType.OpenAI, BackendType.ZhiPuAI, BackendType.Mistral, BackendType.Yi, BackendType.Gemini, BackendType.DeepSeek):
        tool_call_entry = {
            "id": tool_call_id,
            "type": "function",
            "function": {
                "name": tool_name,
                "arguments": tool_arguments,
            },
        }
        if extra_content:
            tool_call_entry["extra_content"] = extra_content
        if tool_cache_control:
            tool_call_entry["cache_control"] = tool_cache_control
        tool_call_message = {
            "content": content,
            "role": "assistant",
            "tool_calls": [tool_call_entry],
        }
    elif backend == BackendType.Anthropic:
        tool_use_block = {
            "type": "tool_use",
            "id": tool_call_id,
            "name": tool_name,
            "input": selected_workflow.get("params", {}),
        }
        # Preserve cache_control for tool_use block
        if tool_cache_control:
            tool_use_block["cache_control"] = tool_cache_control
        tool_call_message = {
            "role": "assistant",
            "content": [tool_use_block],
        }
        if content:
            tool_call_message["content"].insert(0, {"type": "text", "text": content})
    else:
        tool_call_message = {
            "content": json.dumps(
                {
                    "name": message["metadata"]["selected_workflow"]["function_name"],
                    "arguments": json.dumps(message["metadata"]["selected_workflow"]["params"], ensure_ascii=False),
                },
                ensure_ascii=False,
            ),
            "role": "assistant",
        }
    formatted_messages.append(tool_call_message)

    # 工具调用结果消息
    if backend in (BackendType.OpenAI, BackendType.ZhiPuAI, BackendType.Mistral, BackendType.Yi, BackendType.Gemini, BackendType.DeepSeek):
        tool_call_result_message = {
            "role": "tool",
            "tool_call_id": tool_call_id,
            "name": tool_name,
            "content": message["metadata"].get("workflow_result", ""),
        }
    elif backend == BackendType.Anthropic:
        tool_result_block = {
            "type": "tool_result",
            "tool_use_id": message["metadata"]["selected_workflow"]["tool_call_id"],
            "content": message["metadata"].get("workflow_result", ""),
        }
        tool_call_result_message = {
            "role": "user",
            "content": [tool_result_block],
        }
    else:
        tool_call_result_message = {
            "role": "user",
            "content": json.dumps(
                {
                    "function": message["metadata"]["selected_workflow"]["function_name"],
                    "result": message["metadata"].get("workflow_result", ""),
                },
                ensure_ascii=False,
            ),
        }
    formatted_messages.append(tool_call_result_message)

    # if content and backend not in (BackendType.Mistral, BackendType.Anthropic):
    #     formatted_messages.append({"role": "assistant", "content": content})

    return formatted_messages


def transform_from_openai_message(
    message: ChatCompletionMessageParam,
    backend: BackendType,
    native_multimodal: bool = False,
    function_call_available: bool = False,
    process_image: bool = True,
):
    message_dict = cast(dict[str, Any], message)
    role = message.get("role", "user")
    content = message.get("content", "")
    tool_calls = message.get("tool_calls", [])

    if backend == BackendType.Anthropic:
        if role == "tool":
            tool_result_block = {
                "type": "tool_result",
                "tool_use_id": message_dict.get("tool_call_id"),
                "content": message_dict.get("content"),
            }
            # Preserve cache_control if present in message
            if "cache_control" in message_dict:
                tool_result_block["cache_control"] = message_dict["cache_control"]
            return {
                "role": "user",
                "content": [tool_result_block],
            }
        if isinstance(content, list):
            formatted_content = []
            for item in content:
                if isinstance(item, str):
                    formatted_content.append({"type": "text", "text": item})
                elif isinstance(item, dict) and "type" in item:
                    if item["type"] == "image_url":
                        # Extract cache_control from image item if present
                        item_cache_control = item.get("cache_control")
                        formatted_content.append(
                            format_image_message(
                                item["image_url"]["url"],
                                backend,
                                process_image,
                                cache_control=item_cache_control,
                            )
                        )
                    else:
                        # Preserve the item as-is (including cache_control)
                        formatted_content.append(item)
            if tool_calls:
                for tool_call in tool_calls:
                    if isinstance(tool_call, dict) and "function" in tool_call:
                        tool_call_dict = cast(dict[str, Any], tool_call)
                        function_info = tool_call_dict["function"]
                        if isinstance(function_info, dict):
                            tool_use_block = {
                                "type": "tool_use",
                                "id": tool_call_dict["id"],
                                "name": function_info["name"],
                                "input": json.loads(function_info["arguments"]) if function_info.get("arguments") else {},
                            }
                            # Preserve cache_control from tool_call if present
                            if "cache_control" in tool_call_dict:
                                tool_use_block["cache_control"] = tool_call_dict["cache_control"]
                            formatted_content.append(tool_use_block)
            return {"role": role, "content": formatted_content}
        else:
            if tool_calls:
                formatted_content = [{"type": "text", "text": content}] if content else []
                for tool_call in tool_calls:
                    if isinstance(tool_call, dict) and "function" in tool_call:
                        tool_call_dict = cast(dict[str, Any], tool_call)
                        function_info = tool_call_dict["function"]
                        if isinstance(function_info, dict):
                            tool_use_block = {
                                "type": "tool_use",
                                "id": tool_call_dict["id"],
                                "name": function_info["name"],
                                "input": json.loads(function_info["arguments"]) if function_info.get("arguments") else {},
                            }
                            # Preserve cache_control from tool_call if present
                            if "cache_control" in tool_call_dict:
                                tool_use_block["cache_control"] = tool_call_dict["cache_control"]
                            formatted_content.append(tool_use_block)
                return {"role": role, "content": formatted_content}
            else:
                return {"role": role, "content": content}
    else:
        if isinstance(content, str):
            if not function_call_available:
                if message["role"] == "tool":
                    return {
                        "role": "user",
                        "content": f"Tool<tool_call_id: {message['tool_call_id']}> Call result: {content}",
                    }
                elif message["role"] == "assistant":
                    if tool_calls:
                        tool_calls_content = json.dumps(tool_calls, ensure_ascii=False)
                        content += f"Tool calls: {tool_calls_content}"
                    return {"role": role, "content": content}

            return message
        elif isinstance(content, list):
            formatted_content = []
            for item in content:
                if item["type"] == "image_url" and not native_multimodal:
                    formatted_content.append({"type": "text", "text": f"Image<image_url: {item['image_url']['url']}>"})
                else:
                    # Preserve item as-is (including cache_control)
                    formatted_content.append(item)

            if not function_call_available:
                if message["role"] == "tool":
                    role = "user"
                    formatted_content = [
                        {
                            "type": "text",
                            "text": f"Tool<tool_call_id: {message['tool_call_id']}> Call result: {formatted_content}",
                        }
                    ]
                elif message["role"] == "assistant":
                    if tool_calls:
                        tool_calls_content = json.dumps(tool_calls, ensure_ascii=False)
                        formatted_content.append({"type": "text", "text": f"Tool calls: {tool_calls_content}"})

                return {"role": role, "content": formatted_content}

            if tool_calls:
                return {"role": role, "content": formatted_content, "tool_calls": tool_calls}
            else:
                return {"role": role, "content": formatted_content}
        else:
            return message


def format_messages(
    messages: Iterable[ChatCompletionMessageParam | VectorVeinMessage],
    backend: BackendType = BackendType.OpenAI,
    native_multimodal: bool = False,
    function_call_available: bool = False,
    process_image: bool = True,
) -> list:
    """将 VectorVein 和 OpenAI 的 Message 序列化后的格式转换为不同模型支持的格式

    Args:
        messages (list): VectorVein Or OpenAI messages list.
        backend (str, optional): Messages format target backend. Defaults to BackendType.OpenAI.
        native_multimodal (bool, optional): Use native multimodal ability. Defaults to False.
        function_call_available (bool, optional): Use function call ability. Defaults to False.
        process_image (bool, optional): Process image. Defaults to True.

    Returns:
        list: 转换后的消息列表
    """

    def is_vectorvein_message(message):
        return "content_type" in message

    formatted_messages = []

    for message in messages:
        if is_vectorvein_message(message):
            vectorvein_message = cast(dict[str, Any], message)
            match vectorvein_message.get("content_type"):
                case "TXT":
                    pass
                case "WKF":
                    pass
                case _:
                    content_type = vectorvein_message.get("content_type", "unknown")
                    raise ValueError(f"Unsupported message type: {content_type}")
            # 处理 VectorVein 格式的消息
            content_dict = vectorvein_message.get("content")
            if isinstance(content_dict, dict) and "text" in content_dict:
                content = cast(str, content_dict["text"])
            else:
                content = ""
            if vectorvein_message.get("content_type") == "TXT":
                role = "user" if vectorvein_message.get("author_type") == "U" else "assistant"
                formatted_message = format_text_message(content, role, vectorvein_message.get("attachments", []), backend, native_multimodal, process_image)
                formatted_messages.append(formatted_message)
            elif vectorvein_message.get("content_type") == "WKF" and vectorvein_message.get("status") in ("S", "R"):
                formatted_messages.extend(format_workflow_messages(cast(VectorVeinWorkflowMessage, vectorvein_message), content, backend))
        else:
            # 处理 OpenAI 格式的消息
            message = cast(ChatCompletionMessageParam, message)
            formatted_message = transform_from_openai_message(message, backend, native_multimodal, function_call_available, process_image)
            formatted_messages.append(formatted_message)

    return formatted_messages


def format_text_message(
    content: str,
    role: str,
    attachments: list,
    backend: BackendType,
    native_multimodal: bool,
    process_image: bool = True,
    cache_control: dict | None = None,
):
    images_extensions = ("jpg", "jpeg", "png", "bmp")
    has_images = any(attachment.lower().endswith(images_extensions) for attachment in attachments)

    if attachments:
        content += "\n# Attachments:\n"
        content += "\n".join([f"- {attachment}" for attachment in attachments])

    if native_multimodal and has_images:
        text_block = {"type": "text", "text": content}
        # Add cache_control to the text block if provided
        if cache_control is not None:
            text_block["cache_control"] = cache_control
        return {
            "role": role,
            "content": [
                text_block,
                *[
                    format_image_message(image=attachment, backend=backend, process_image=process_image)
                    for attachment in attachments
                    if attachment.lower().endswith(images_extensions)
                ],
            ],
        }
    else:
        # If cache_control is provided, we need to use list format for content
        if cache_control is not None:
            text_block = {"type": "text", "text": content, "cache_control": cache_control}
            return {"role": role, "content": [text_block]}
        return {"role": role, "content": content}


def generate_tool_use_system_prompt(tools: list | str, format_type: str = "json") -> str:
    if format_type == "json":
        return (
            "You have access to the following tools. Use them if required and wait for the tool call result. Stop output after calling a tool.\n\n"
            f"# Tools\n{tools}\n\n"
            "# Requirements when using tools\n"
            "Must starts with <|▶|> and ends with <|◀|>\n"
            "Must be valid JSON format and pay attention to escape characters.\n"
            '## Output format\n<|▶|>{"name": "<function name:str>", "arguments": <arguments:dict>}<|◀|>\n\n'
            '## Example output\n<|▶|>{"name": "get_current_weather", "arguments": {"location": "San Francisco, CA"}}<|◀|>'
        )
    elif format_type == "xml":
        return (
            "You have access to the following tools. Use them if required and wait for the tool call result. Stop output after calling a tool.\n\n"
            f"# Tools\n{tools}\n\n"
            "# Requirements when using tools\n"
            "Must starts with <|▶|> and ends with <|◀|>\n"
            "Must be valid XML format.\n"
            "## Output format\n<|▶|><invoke><tool_name>[function name:str]</tool_name><parameters><parameter_1_name>[parameter_1_value]</parameter_1_name><parameter_2_name>[parameter_2_value]</parameter_2_name>...</parameters></invoke><|◀|>\n\n"
            "## Example output\n<|▶|><invoke><tool_name>calculator</tool_name><parameters><first_operand>1984135</first_operand><second_operand>9343116</second_operand><operator>*</operator></parameters></invoke><|◀|>"
        )
    else:
        return ""



from __future__ import annotations

import httpx
from typing import TYPE_CHECKING

from .base_client import EmbeddingClient, AsyncEmbeddingClient
from ..types.enums import EmbeddingBackendType

if TYPE_CHECKING:
    from ..settings import Settings
    from ..types.settings import SettingsDict


def create_embedding_client(
    backend: EmbeddingBackendType | str,
    model: str | None = None,
    random_endpoint: bool = True,
    endpoint_id: str = "",
    http_client: httpx.Client | None = None,
    settings: Settings | SettingsDict | None = None,
) -> EmbeddingClient:
    return EmbeddingClient(
        backend=backend,
        model=model,
        random_endpoint=random_endpoint,
        endpoint_id=endpoint_id,
        http_client=http_client,
        settings=settings,
    )


def create_async_embedding_client(
    backend: EmbeddingBackendType | str,
    model: str | None = None,
    random_endpoint: bool = True,
    endpoint_id: str = "",
    http_client: httpx.AsyncClient | None = None,
    settings: Settings | SettingsDict | None = None,
) -> AsyncEmbeddingClient:
    return AsyncEmbeddingClient(
        backend=backend,
        model=model,
        random_endpoint=random_endpoint,
        endpoint_id=endpoint_id,
        http_client=http_client,
        settings=settings,
    )


__all__ = [
    "EmbeddingBackendType",
    "EmbeddingClient",
    "AsyncEmbeddingClient",
    "create_embedding_client",
    "create_async_embedding_client",
]
